﻿using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using RecrownedAthenaeum.Camera;
using System;
using System.Collections.Generic;

namespace RecrownedAthenaeum.Render
{
    /// <summary>
    /// A batch used to draw primitive shapes by batching together vertices.
    /// </summary>
    public class PrimitiveBatch : IDisposable
    {
        VertexPositionColor[] vertices;
        int bufferPosition;
        int maxVertices;

        BasicEffect basicEffect;
        PrimitiveType primitiveType;

        int verticesPerPrimitive;
        GraphicsDevice graphicsDevice;
        bool began;
        bool disposed;
        Camera3D camera;

        /// <summary>
        /// Creates a batch used to draw primitives.
        /// </summary>
        /// <param name="camera">The current camera being used. Will use default set in <see cref="Configuration"/> if left null.</param>
        /// <param name="graphicsDevice">The graphics device used to draw the primitives. Will be using <see cref="Configuration"/>'s graphics device from graphics device manager if null. Default is null.</param>
        /// <param name="verticesPerBatch">The amount of vertices every batch can hold before flushing. Default is 450. Should be changed to be the most optimal number if possible to prevent unnecessary resizing. Especially if using strip primitive types.</param>
        public PrimitiveBatch(Camera2D camera = null, GraphicsDevice graphicsDevice = null, int verticesPerBatch = 500)
        {
            this.graphicsDevice = graphicsDevice ?? (Configuration.GraphicsDeviceManager.GraphicsDevice);
            this.camera = camera ?? (Configuration.Camera2D);
            basicEffect = new BasicEffect(this.graphicsDevice);
            basicEffect.VertexColorEnabled = true;
            vertices = new VertexPositionColor[verticesPerBatch+1];
            maxVertices = verticesPerBatch;
        }

        /// <summary>
        /// Starts the batch. Batch cannot be started twice.
        /// </summary>
        /// <param name="primitiveType">The type of primitive this batch would be drawing.</param>
        public void Begin(PrimitiveType primitiveType)
        {
            if (began) throw new InvalidOperationException("Begin is being called twice before being ended.");
            if (disposed) throw new ObjectDisposedException(this.GetType().Name);
            this.primitiveType = primitiveType;
            verticesPerPrimitive = 0;
            switch (primitiveType)
            {
                case PrimitiveType.LineList: verticesPerPrimitive = 2; break;
                case PrimitiveType.TriangleList: verticesPerPrimitive = 3; break;
                default: verticesPerPrimitive = 1; break;
            }

            basicEffect.World = camera.worldMatrix;
            basicEffect.View = camera.ViewMatrix;
            basicEffect.Projection = camera.projectionMatrix;
            basicEffect.CurrentTechnique.Passes[0].Apply();
            began = true;
        }

        /// <summary>
        /// Ends the batch. Begin needs to be called before end.
        /// </summary>
        public void End()
        {
            if (disposed) throw new ObjectDisposedException(this.GetType().Name);
            if (!began) throw new InvalidOperationException("Begin must be called before ending.");
            Flush();
            began = false;
        }

        /// <summary>
        /// Adds a vertex position for the primitive being drawn. The batch needs to have beens started before this.
        /// </summary>
        /// <param name="vertex">The vector that represents the vertex.</param>
        /// <param name="color">The color of that vertex.</param>
        public void AddVertex(Vector2 vertex, Color color)
        {
            if (!began) throw new InvalidOperationException("Begin needs to be called before adding vertex.");
            if (disposed) throw new ObjectDisposedException(this.GetType().Name);
            if (primitiveType != PrimitiveType.LineStrip && primitiveType != PrimitiveType.TriangleStrip)
            {
                if (bufferPosition + verticesPerPrimitive > maxVertices && (bufferPosition % maxVertices == 0))
                {
                    Flush();
                }
                else
                {
                    throw new InvalidOperationException("Buffer size doesn't match with primitive type.");
                }
            }
            else if (bufferPosition + verticesPerPrimitive > maxVertices)
            {
                throw new InvalidOperationException("Buffer size isn't large enough.");
            }

            vertices[bufferPosition] = new VertexPositionColor(new Vector3(vertex, 0), color);
            bufferPosition++;
        }

        /// <summary>
        /// Flushes the batch. Automatically called if required if using <see cref="PrimitiveType.LineList"/> or <see cref="PrimitiveType.TriangleList"/>. Otherwise, manual flushing is required.
        /// </summary>
        public void Flush()
        {
            if (!began) throw new InvalidOperationException("Begin needs to be called before flushing.");
            if (disposed) throw new ObjectDisposedException(this.GetType().Name);
            if (bufferPosition == 0) return;
            graphicsDevice.DrawUserPrimitives(primitiveType, vertices, 0, bufferPosition / verticesPerPrimitive);
            bufferPosition = 0;
        }

        /// <summary>
        /// Disposes this.
        /// </summary>
        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        /// <summary>
        /// Overridable dispose.
        /// </summary>
        /// <param name="disposing">True for when user called for this to be disposed. False otherwise.</param>
        public virtual void Dispose(bool disposing)
        {
            if (disposed) throw new ObjectDisposedException(this.GetType().Name);
            disposed = true;
            if (disposing && !disposed)
            {
                basicEffect.Dispose();
            }
        }

        /// <summary>
        /// Destructor.
        /// </summary>
        ~PrimitiveBatch()
        {
            Dispose(false);
        }
    }
}
